// clear console and variables

clear
clc

exec("networkOptimizer_functions.sci",0);
exec('optimizations/optObjectives_functions.sci',0);

mprintf("\n***************************\n");
mprintf("* NETWORK OPTIMIZER       *\n");
mprintf("***************************\n");

mprintf("\nChoose the optimization objective:\n");
mprintf("\n1) Economic cost minimization");
mprintf("\n2) Power consumption minimization");
mprintf("\n3) Transmission Delay minimization");
mprintf("\n4) Error rate minimization\n\n");

opt = mscanf("%d");

mprintf("\nDigit the name of the xml file containing the network description\n\n");

nwFile_name = mscanf("%s");

mprintf("\nDigit the name of the xml file containing the network technological library\n\n");

tlFile_name = mscanf("%s");

nwXMLFile_name = "inputs/" + nwFile_name + ".xml";
techLibXMLFile_name = "inputs/" + tlFile_name + ".techlib.xml";

[T,F,N,A,Z,C] = getNetwork(nwXMLFile_name);
network = struct('tasks',T,'dataFlows',F,'nodes',N,'abstractChannels',A,'zones',Z,'contiguities',C);
[N_lib,A_lib] = getTechLib(techLibXMLFile_name);
//[N_lib,A_lib] = getTechLib();
techLib = struct('nodes',N_lib,'abstractChannels',A_lib);
Log = %t;
obj = "";
manFile_name ="";

if (opt == 1) then
    obj = "Economic cost";
    f = cost(list(network,network));
    mprintf("\nECONOMIC COST MINIMIZATION - %s\n",nwFile_name);
    start = tic();
    //[manipulations_opt, results_opt, intResults] = minCost(network,Log);
    [manipulations_opt, results_opt] = minCost(network,Log);
    stop = toc();
    manFile_name = "outputs/" + nwFile_name + "_minCost.man";
    //optFile_name = "outputs/" + nwFile_name + "_minCost.opt";
end
if (opt == 2) then
    obj = "Power consumption";
    f = powerConsumption(list(network,network));
    mprintf("\nPOWER CONSUMPTION MINIMIZATION - %s\n",nwFile_name);
    start = tic();
    //[manipulations_opt, results_opt, intResults] = minPower(network,Log);
    [manipulations_opt, results_opt] = minPower(network,techLib,Log);
    stop = toc();
    manFile_name = "outputs/" + nwFile_name + "_minPower.man";
//    optFile_name = "outputs/" + nwFile_name + "_minPower.opt";
end
if (opt == 3) then
    obj = "Transmission delay";
    f = delay(list(network,network));
    mprintf("\nTRANSMISSION DELAY MINIMIZATION - %s\n",nwFile_name);
    start = tic();
    //[manipulations_opt, results_opt, intResults] = minDelay(network,Log);
    [manipulations_opt, results_opt] = minDelay(network,techLib,Log);
    stop = toc();
    manFile_name = "outputs/" + nwFile_name + "_minDelay.man";
    //optFile_name = "outputs/" + nwFile_name + "_minDelay.opt";
end
if (opt == 4) then
    obj = "Error rate";
    f = errorRate(list(network,network));
    mprintf("\nERROR RATE MINIMIZATION - %s\n",nwFile_name);
    start = tic();
    //[manipulations_opt, results_opt, intResults] = minErrorRate(network,Log);
    [manipulations_opt, results_opt] = minErrorRate(network,techLib,Log);
    stop = toc();
    manFile_name = "outputs/" + nwFile_name + "_minErrorRate.man";
    //optFile_name = "outputs/" + nwFile_name + "_minErrorRate.opt";
end

moveTasks = manipulations_opt.moveTasks;
moveDataFlows = manipulations_opt.moveDataFlows;
removeDataFlows = manipulations_opt.removeDataFlows;
addNodes = manipulations_opt.addNodes;
removeNodes = manipulations_opt.removeNodes;
addChannels = manipulations_opt.addChannels;
moveChannels = manipulations_opt.moveChannels;
removeChannels = manipulations_opt.removeChannels;

f_opt = results_opt.f_opt;

if (Log) then
    mprintf("\n-----------------------------------------------------------------------");
    mprintf("\nINITIAL VALUE OF THE OBJECTIVE FUNCTION");
    mprintf("\n-----------------------------------------------------------------------\n");
    mprintf("\n%s = %f\n",obj,f);
    mprintf("\n-----------------------------------------------------------------------");
    mprintf("\nFINAL VALUE OF THE OBJECTIVE FUNCTION");
    mprintf("\n-----------------------------------------------------------------------\n");
    mprintf("\n%s = %f\n",obj,f_opt);
    
    mprintf("\n-----------------------------------------------------------------------");
    mprintf("\nMANIPULATION RULES");
    mprintf("\n-----------------------------------------------------------------------\n");
    
    if (length(moveTasks) > 0) then
        mprintf("\n> MOVE TASK ( task : source_node --> destination_node )\n");
        for i=1:length(moveTasks)
            mprintf("\n\t%s : %s --> %s\n",moveTasks(i).task,moveTasks(i).srcNode,moveTasks(i).dstNode);
        end
    end
    
    if (length(moveDataFlows) > 0) then
        mprintf("\n> MOVE DATA FLOW ( data_flow : source_channel --> destination_channel )\n");
        for i=1:length(moveDataFlows)
            mprintf("\n\t%s : %s --> %s\n",moveDataFlows(i).dataFlow,moveDataFlows(i).srcChannel,moveDataFlows(i).dstChannel);
        end
    end
    
    if (length(removeDataFlows) > 0) then
        mprintf("\n> REMOVE DATA FLOW ( data_flow (channel) )\n");
        for i=1:length(removeDataFlows)
            mprintf("\n\t%s (%s)\n",removeDataFlows(i).dataFlow,removeDataFlows(i).channel);
        end
    end
    
    if (length(addNodes) > 0) then
        mprintf("\n> ADD NODE ( node : [lib_node] (zone) )\n");
        for i=1:length(addNodes)
            mprintf("\n\t%s [%s] (%s)\n",addNodes(i).node,addNodes(i).libNode,addNodes(i).zone);
        end
    end
    
    if (length(removeNodes) > 0) then
        mprintf("\n> REMOVE NODE ( node (zone) )\n");
        for i=1:length(removeNodes)
            mprintf("\n\t%s (%s)\n",removeNodes(i).node,removeNodes(i).zone);
        end
    end
    
    if (length(addChannels) > 0) then
        mprintf("\n> ADD CHANNEL ( channel [lib_channel] )\n");
        for i=1:length(addChannels)
            mprintf("\n\t%s [%s]\n",addChannels(i).channel,addChannels(i).libChannel);
        end
    end
    
    if (length(moveChannels) > 0) then
        mprintf("\n> MOVE CHANNEL ( channel : old_node --> new_node )\n");
        for i=1:length(moveChannels)
            mprintf("\n\t%s : %s --> %s\n",moveChannels(i).channel,moveChannels(i).old_node,moveChannels(i).new_node);
        end
    end
    
    if (length(removeChannels) > 0) then
        mprintf("\n> REMOVE CHANNEL ( channel )\n");
        for i=1:length(removeChannels)
            mprintf("\n\t%s\n",removeChannels(i).channel);
        end
    end
end
    
mprintf("\n-----------------------------------------------------------------------");
mprintf("\nOptimization time: %f sec",stop);
mprintf("\n-----------------------------------------------------------------------\n");

if (manFile_name <> "") then
    writeManipulations(manFile_name,techLibXMLFile_name,manipulations_opt);
end
