// FUNCTIONS

function e = getElement(E,e_name)
    for i=1:length(E)
        if (E(i).name == e_name) then
            e = E(i);
        end
    end
endfunction

// Get Network definition from a XML file
function [T,F,N,A,Z,C] = getNetwork(nwFileName)    
    if (~isdef("nwFileName","local")) then
        error(sprintf(gettext("%s: nwFileName is mandatory"),"getNetwork"));
    end 
    
    // NETWORK DEFINITION

    T = list();
    F = list();
    N = list();
    A = list();
    Z = list();
    C = list();
    
    network = xmlRead(nwFileName);
    
    for i=1:network.root.children.size
        XMLElement_name = network.root.children(i).name;
        
        // Tasks
        
        if (XMLElement_name == "TASKS") then
            tasks = network.root.children(i);
            
            for j=1:tasks.children.size
                task = tasks.children(j);
                task_name = task.attributes.name;
                task_mobility = strtod(task.attributes.m);
                
                task_capacityList = list();
                
                for k=1:task.children.size
                    task_child = task.children(k);
                    if (task_child.name == "CAPACITY") then
                        task_capacityList($+1) = strtod(task_child.attributes.value);
                    end
                end
                
                task_capacity = zeros(1,length(task_capacityList));
                
                for k=1:length(task_capacityList)
                    task_capacity(1,k) = task_capacityList(k);
                end
                
                t = struct('name',task_name,'c',task_capacity,'m',task_mobility);
                T($+1) = t;
            end
        end
        
        // Data Flows
        
        if (XMLElement_name == "DATAFLOWS") then
            dataFlows = network.root.children(i);
            
            for j=1:dataFlows.children.size
                dataFlow = dataFlows.children(j);
                dataFlow_name = dataFlow.attributes.name;
                
                for k=1:dataFlow.children.size
                    dataFlow_child = dataFlow.children(k);
                    if (dataFlow_child.name == "TS") then
                        ts_name = dataFlow_child.attributes.name;
                    end
                    if (dataFlow_child.name == "TD") then
                        td_name = dataFlow_child.attributes.name;
                    end
                    if (dataFlow_child.name == "CAPACITY") then
                        maxThroughput = strtod(dataFlow_child.attributes.max_throughput);
                        maxDelay = strtod(dataFlow_child.attributes.max_delay);
                        maxErrorRate = strtod(dataFlow_child.attributes.max_error_rate);
                        dataFlow_capacity = struct('max_throughput',maxThroughput,'max_delay',maxDelay,'max_error_rate',maxErrorRate);
                    end
                end
                
                f = struct('name',dataFlow_name,'ts',getElement(T,ts_name),'td',getElement(T,td_name),'c',dataFlow_capacity);
                F($+1) = f;
            end     
        end
        
        // Nodes
        
        if (XMLElement_name == "NODES") then
            nodes = network.root.children(i);
            
            for j=1:nodes.children.size
                node = nodes.children(j);
                node_name = node.attributes.name;
                node_powerBudget = strtod(node.attributes.p);
                node_cost = strtod(node.attributes.k);
                node_mobility = strtod(node.attributes.m);
                
                t_list = list();
                node_capacityList = list();
                node_gammaList = list();
                
                for k=1:node.children.size
                    node_child = node.children(k);
                    if (node_child.name == "T") then
                        for z=1:node_child.children.size
                            task_name = node_child.children(z).attributes.name;
                            t_list($+1) = getElement(T,task_name);
                        end
                    end
                    if (node_child.name == "CAPACITY") then
                        node_capacityList($+1) = strtod(node_child.attributes.value);
                        node_gammaList($+1) = strtod(node_child.attributes.gamma);
                    end
                end
                
                node_capacity = zeros(1,length(node_capacityList));
                node_gamma = zeros(1,length(node_gammaList));
                
                for k=1:length(node_capacityList)
                    node_capacity(1,k) = node_capacityList(k);
                    node_gamma(1,k) = node_gammaList(k);
                end
                
                n = struct('name',node_name,'t',t_list,'c',node_capacity,'p',node_powerBudget,'k',node_cost,'g',node_gamma,'m',node_mobility);
                N($+1) = n;
            end
        end
        
        // Abstract Channels
        
        if (XMLElement_name == "ABSTRACTCHANNELS") then
            abstractChannels = network.root.children(i);
            
            for j=1:abstractChannels.children.size
                abstractChannel = abstractChannels.children(j);
                abstractChannel_name = abstractChannel.attributes.name;
                abstractChannel_power = strtod(abstractChannel.attributes.p);
                abstractChannel_cost = strtod(abstractChannel.attributes.k);
                abstractChannel_wireless = strtod(abstractChannel.attributes.w);
                
                n_list = list();
                f_list = list();
                
                for k=1:abstractChannel.children.size
                    abstractChannel_child = abstractChannel.children(k);
                    if (abstractChannel_child.name == "N") then
                        for z=1:abstractChannel_child.children.size
                            node_name = abstractChannel_child.children(z).attributes.name;
                            n_list($+1) = getElement(N,node_name);
                        end
                    end
                    if (abstractChannel_child.name == "F") then
                        for z=1:abstractChannel_child.children.size
                            dataFlow_name = abstractChannel_child.children(z).attributes.name;
                            f_list($+1) = getElement(F,dataFlow_name);
                        end
                    end
                    if (abstractChannel_child.name == "CAPACITY") then
                        throughput = strtod(abstractChannel_child.attributes.throughput);
                        delay = strtod(abstractChannel_child.attributes.delay);
                        errorRate = strtod(abstractChannel_child.attributes.error_rate);
                        abstractChannel_capacity = struct('throughput',throughput,'delay',delay,'error_rate',errorRate);
                    end
                end
                
                a = struct('name',abstractChannel_name,'n',n_list,'f',f_list,'c',abstractChannel_capacity,'p',abstractChannel_power,'k',abstractChannel_cost,'w',abstractChannel_wireless);
                A($+1) = a;
            end
        end
        
        // Zones
        
        if (XMLElement_name == "ZONES") then
            zones = network.root.children(i);
            
            for j=1:zones.children.size
                zone = zones.children(j);
                zone_name = zone.attributes.name;
                
                n_list = list();
                
                for k=1:zone.children.size
                    zone_child = zone.children(k);
                    if (zone_child.name == "N") then
                        for z=1:zone_child.children.size
                            node_name = zone_child.children(z).attributes.name;
                            n_list($+1) = getElement(N,node_name);
                        end
                    end
                end
                
                z = struct('name',zone_name,'n',n_list);
                Z($+1) = z;
            end
        end
        
        // Contiguities
        
        if (XMLElement_name == "CONTIGUITIES") then
            contiguities = network.root.children(i);
            
            for j=1:contiguities.children.size
                contiguity = contiguities.children(j);
                contiguity_name = contiguity.attributes.name;
                
                for k=1:contiguity.children.size
                    contiguity_child = contiguity.children(k);
                    if (contiguity_child.name == "Z1") then
                        z1_name = contiguity_child.attributes.name;
                    end
                    if (contiguity_child.name == "Z2") then
                        z2_name = contiguity_child.attributes.name;
                    end
                    if (contiguity_child.name == "R") then
                        addedDelay = strtod(contiguity_child.attributes.added_delay);
                        addedErrorRate = strtod(contiguity_child.attributes.added_error_rate);
                        throughputRedFactor = strtod(contiguity_child.attributes.throughput_red_factor);
                        powerRedFactor = strtod(contiguity_child.attributes.power_red_factor);
                        resistance = struct('added_delay',addedDelay,'added_error_rate',addedErrorRate,'throughput_red_factor',throughputRedFactor,'power_red_factor',powerRedFactor);
                    end
                end
                
                c = struct('name',contiguity_name,'z1',getElement(Z,z1_name),'z2',getElement(Z,z2_name),'r',resistance);
                C($+1) = c;
            end
        end
        
    end
endfunction

function [N,A] = getTechLib(techLibFileName)    
    if (~isdef("techLibFileName","local")) then
        N = list();
        A = list();
        return
    end 
    
    N = list();
    A = list();
    
    techLib = xmlRead(techLibFileName);
    
    for i=1:techLib.root.children.size
        XMLElement_name = techLib.root.children(i).name;
         
        // Nodes
        
        if (XMLElement_name == "NODES") then
            nodes = techLib.root.children(i);
            
            for j=1:nodes.children.size
                node = nodes.children(j);
                node_name = node.attributes.name;
                node_powerBudget = strtod(node.attributes.p);
                node_cost = strtod(node.attributes.k);
                node_mobility = strtod(node.attributes.m);
                
                node_capacityList = list();
                node_gammaList = list();
                
                for k=1:node.children.size
                    node_child = node.children(k);
                    if (node_child.name == "CAPACITY") then
                        node_capacityList($+1) = strtod(node_child.attributes.value);
                        node_gammaList($+1) = strtod(node_child.attributes.gamma);
                    end
                end
                
                node_capacity = zeros(1,length(node_capacityList));
                node_gamma = zeros(1,length(node_gammaList));
                
                for k=1:length(node_capacityList)
                    node_capacity(1,k) = node_capacityList(k);
                    node_gamma(1,k) = node_gammaList(k);
                end
                
                n = struct('instances',0,'name',node_name,'t',list(),'c',node_capacity,'p',node_powerBudget,'k',node_cost,'g',node_gamma,'m',node_mobility);
                N($+1) = n;
            end
        end
        
        // Abstract Channels
        
        if (XMLElement_name == "ABSTRACTCHANNELS") then
            abstractChannels = techLib.root.children(i);
            
            for j=1:abstractChannels.children.size
                abstractChannel = abstractChannels.children(j);
                abstractChannel_name = abstractChannel.attributes.name;
                abstractChannel_power = strtod(abstractChannel.attributes.p);
                abstractChannel_cost = strtod(abstractChannel.attributes.k);
                abstractChannel_wireless = strtod(abstractChannel.attributes.w);
                
                for k=1:abstractChannel.children.size
                    abstractChannel_child = abstractChannel.children(k);
                    if (abstractChannel_child.name == "CAPACITY") then
                        throughput = strtod(abstractChannel_child.attributes.throughput);
                        delay = strtod(abstractChannel_child.attributes.delay);
                        errorRate = strtod(abstractChannel_child.attributes.error_rate);
                        abstractChannel_capacity = struct('throughput',throughput,'delay',delay,'error_rate',errorRate);
                    end
                end
                
                a = struct('instances',0,'name',abstractChannel_name,'n',list(),'f',list(),'c',abstractChannel_capacity,'p',abstractChannel_power,'k',abstractChannel_cost,'w',abstractChannel_wireless);
                A($+1) = a;
            end
        end
    end
    
endfunction


function writeManipulations(manFile_name,techLibFile_name,manipulations_opt)
    manXMLFile_name = manFile_name + ".xml";
    manXMLFile = xmlDocument(manFile_name);
    
    manXMLFile = xmlReadStr("<MANIPULATIONS></MANIPULATIONS>");
    
    moveTasks = manipulations_opt.moveTasks;
    moveDataFlows = manipulations_opt.moveDataFlows;
    removeDataFlows = manipulations_opt.removeDataFlows;
    removeNodes = manipulations_opt.removeNodes;
    moveChannels = manipulations_opt.moveChannels;
    removeChannels = manipulations_opt.removeChannels;
    
    // Write nodes and channels from the technological
    // library used in the manipulations   

    libNodes = list();
    libChannels = list();
    
    for i=1:length(addNodes)
        libNode = addNodes(i).libNode;
        isInList = %f;
        for j=1:length(libNodes)
            if (libNodes(j) == libNode) then
                isInList = %t;
                break
            end
        end
        if (~isInList) then
            libNodes($+1) = libNode;
        end
    end
    
    for i=1:length(addChannels)
        libChannel = addChannels(i).libChannel;
        isInList = %f;
        for j=1:length(libChannels)
            if (libChannels(j) == libChannel) then
                isInList = %t;
                break
            end
        end
        if (~isInList) then
            libChannels($+1) = libChannel;
        end
    end
    
    techLib = xmlRead(techLibFile_name);
    
    XML_libNodes = list();
    XML_libChannels = list();
        
    for i=1:techLib.root.children.size
        XMLElement_name = techLib.root.children(i).name;
         
        // Nodes
        
        if (XMLElement_name == "NODES") then
            nodes = techLib.root.children(i);
            
            for j=1:nodes.children.size
                XML_libNodes($+1) = nodes.children(j);
            end
        end
        
        // Abstract Channels
        
        if (XMLElement_name == "ABSTRACTCHANNELS") then
            abstractChannels = techLib.root.children(i);
            
            for j=1:abstractChannels.children.size
                XML_libChannels($+1) = abstractChannels.children(j);
            end
        end
        
    end
    
    techLib_XMLElement = xmlElement(manXMLFile,"TECHLIB");      

    for i=1:length(libNodes)
        for j=1:length(XML_libNodes)
            if (libNodes(i) == XML_libNodes(j).attributes.name) then
                xmlAppend(techLib_XMLElement,XML_libNodes(j));
            end
        end
    end
    
    for i=1:length(libChannels)
        for j=1:length(XML_libChannels)
            if (libChannels(i) == XML_libChannels(j).attributes.name) then
                xmlAppend(techLib_XMLElement,XML_libChannels(j));
            end
        end
    end
    
    xmlAppend(manXMLFile.root,techLib_XMLElement);
    
    // Write the manipulations
    
    for i=1:length(moveTasks)
        man = moveTasks(i);
        
        taskToMove_XMLElement = xmlElement(manXMLFile,"MOVETASK");
        taskToMove_XMLElement = xmlSetAttributes(taskToMove_XMLElement, ["task" man.task ; "sourceNode" man.srcNode ; "destinationNode" man.dstNode]);
        
        xmlAppend(manXMLFile.root,taskToMove_XMLElement);
    end
    
    for i=1:length(moveDataFlows)
        man = moveDataFlows(i);
        
        dfToMove_XMLElement = xmlElement(manXMLFile,"MOVEDATAFLOW");
        dfToMove_XMLElement = xmlSetAttributes(dfToMove_XMLElement, ["dataFlow" man.dataFlow ; "sourceChannel" man.srcChannel ; "destinationChannel" man.dstChannel]);
        
        xmlAppend(manXMLFile.root,dfToMove_XMLElement);
    end
    
    for i=1:length(removeDataFlows)
        man = removeDataFlows(i);
        
        dfToRemove_XMLElement = xmlElement(manXMLFile,"REMOVEDATAFLOW");
        dfToRemove_XMLElement = xmlSetAttributes(dfToRemove_XMLElement, ["dataFlow" man.dataFlow ; "channel" man.channel]);
        
        xmlAppend(manXMLFile.root,dfToRemove_XMLElement);
    end
    
    for i=1:length(addNodes)
        man = addNodes(i);
        
        addNodes_XMLElement = xmlElement(manXMLFile,"ADDNODE");
        addNodes_XMLElement = xmlSetAttributes(addNodes_XMLElement, ["node" man.node ; "libNode" man.libNode ; "zone" man.zone]);
        
        xmlAppend(manXMLFile.root,addNodes_XMLElement);
    end
    
    for i=1:length(removeNodes)
        man = removeNodes(i);
        
        nodeToRemove_XMLElement = xmlElement(manXMLFile,"REMOVENODE");
        nodeToRemove_XMLElement = xmlSetAttributes(nodeToRemove_XMLElement, ["node" man.node ; "zone" man.zone]);
        
        xmlAppend(manXMLFile.root,nodeToRemove_XMLElement);
    end
    
    for i=1:length(addChannels)
        man = addChannels(i);
        
        chToAdd_XMLElement = xmlElement(manXMLFile,"ADDCHANNEL");
        chToAdd_XMLElement = xmlSetAttributes(chToAdd_XMLElement, ["channel" man.channel ; "libChannel" man.libChannel]);
        
        xmlAppend(manXMLFile.root,chToAdd_XMLElement);
    end
    
    for i=1:length(moveChannels)
        man = moveChannels(i);
        
        chToMove_XMLElement = xmlElement(manXMLFile,"MOVECHANNEL");
        chToMove_XMLElement = xmlSetAttributes(chToMove_XMLElement, ["channel" man.channel ; "oldNode" man.old_node ; "newNode" man.new_node]);
        
        xmlAppend(manXMLFile.root,chToMove_XMLElement);
    end
    
    for i=1:length(removeChannels)
        man = removeChannels(i);
        
        chToRemove_XMLElement = xmlElement(manXMLFile,"REMOVECHANNEL");
        chToRemove_XMLElement = xmlSetAttributes(chToRemove_XMLElement, ["channel" man.channel]);
        
        xmlAppend(manXMLFile.root,chToRemove_XMLElement);
    end
    
    xmlWrite(manXMLFile,manXMLFile_name);
endfunction
