/////////////////////
//  MANIPULATIONS  //
/////////////////////

// manipulation data struct, containing all the possible network manipulations
manipulations = struct('moveTasks',list(),'moveDataFlows',list(),'removeDataFlows',list(),'addNodes',list(),'removeNodes',list(),'addChannels',list(),'moveChannels',list(),'removeChannels',list());

// Move task from a source node to a destination node
function [srcNode,dstNode] = moveTask(srcTask,srcNode,dstNode)
    for i=1:length(srcNode.t)
        if (srcNode.t(i).name == srcTask.name) then
            srcNode.t(i) = null();
            break
        end
    end
    dstNode.t($+1) = srcTask;
endfunction

// Move a data flow from a source channel to a destination channel
function [srcChannel,dstChannel] = moveDataFlow(dataFlow,srcChannel,dstChannel)
    for i=1:length(srcChannel.f)
       if (srcChannel.f(i).name == dataFlow.name)
           srcChannel.f(i) = null();
           break
       end
    end
    dstChannel.f($+1) = dataFlow;
endfunction

// Remove a data flow from a channel and the data flows' list
function [dataFlows,channel] = removeDataFlow(dataFlow,dataFlows,channel)
    for i=1:length(channel.f)
        if (channel.f(i).name == dataFlow.name) then
            channel.f(i) = null();
            break
        end
    end
    for i=1:length(dataFlows)
        if (dataFlows(i).name == dataFlow.name) then
            dataFlows(i) = null();
            break
        end
    end
endfunction

// Remove a node from the nodes' list, the channels and the zone containing it
function [nodes,channels,zones] = removeNode(node,nodes,channels,zones)
    for i=1:length(nodes)
        if (nodes(i).name == node.name) then
            nodes(i) = null();
            break
        end
    end
    for i=1:length(channels)
        for j=1:length(channels(i).n)
            if (channels(i).n(j).name == node.name) then
                channels(i).n(j) = null();
                break
            end
        end
    end
    for i=1:length(zones)
        for j=1:length(zones(i).n)
            if (zones(i).n(j).name == node.name) then
                zones(i).n(j) = null();
                break
            end
        end
    end
endfunction

// Move a channel from a source node to a destination node
function channel = moveChannel(channel,srcNode,dstNode)
    for i=1:length(channel.n)
        if (channel.n(i).name == srcNode.name) then
            channel.n(i) = null();
            break
        end
    end
    
    if (~isElementOf(channel.n,dstNode)) then
        channel.n($+1) = dstNode;
    end
endfunction

// Remove a channel from the channels' list
function channels = removeChannel(channel,channels)
    for i=1:length(channels)
        if (channels(i).name == channel.name) then
            channels(i) = null();
            break
        end
    end
endfunction

/////////////////////////////
//  LIBRARY MANIPULATIONS  //
/////////////////////////////

// Add a node from the technological library
function [nodes,zones] = addNode(node,zone,nodes,zones)
    nodes($+1) = node;
    
    if (zone <> -1) then
        for i=1:length(zones)
            if (zones(i).name == zone.name) then
                zones(i).n($+1) = node;
                break
            end
        end    
    end
endfunction

// Add a channel from the technological library
function channels = addChannel(channel,channels)
    channels($+1) = channel;
endfunction

///////////////////////
//  OTHER FUNCTIONS  //
///////////////////////

// Remove empty channels from the channels' list
function [channels_new,removeChannels] = cleanChannels(channels,removeChannels)        
    channels_new = list();
    for i=1:length(channels)
        if ((length(channels(i).f) == 0) | (length(channels(i).n) == 0)) then
            removeChannels($+1) = struct('channel',channels(i).name);
        else
            channels_new($+1) = channels(i);
        end
    end
endfunction

// Remove empty nodes from the nodes' list, the channels and the zone containing it
function [nodes_new,channels,zones,removeNodes] = cleanNodes(nodes,channels,zones,removeNodes)        
    nodes_new = list();
    for i=1:length(nodes)
        if (length(nodes(i).t) == 0) then
            for j=1:length(channels)
                for k=1:length(channels(j).n)
                    if (channels(j).n(k).name == nodes(i).name) then
                        channels(j).n(k) = null();
                        break
                    end
                end
            end
            zone = getZoneFromNode(nodes(i),zones);
            for j=1:length(zones)
                for k=1:length(zones(j).n)
                    if (zones(j).n(k).name == nodes(i).name) then
                        zones(j).n(k) = null();
                        break
                    end
                end
            end
            if (zone <> -1) then
                removeNodes($+1) = struct('node',nodes(i).name,'zone',zone.name);
            else
                removeNodes($+1) = struct('node',nodes(i).name,'zone',"");
            end
        else
            nodes_new($+1) = nodes(i);
        end
    end
endfunction
