/////////////////////////
//  NETWORK UTILITIES  //
/////////////////////////

// check if an element belongs to a list of elements of the same type
function isE = isElementOf(E,e)
    isE = %f;
    for i=1:length(E)
        if (E(i).name == e.name) then
            isE = %t;
            break;
        end
    end
endfunction

// return the node containing the given task
function node = getNodeFromTask(task,nodes)
    for i=1:length(nodes)
        if (isElementOf(nodes(i).t,task)) then
            node = nodes(i);
            break
        end
    end
endfunction

// return the channel containing the given data flow
function channel = getChannelFromDataFlow(dataFlow,channels)
    for i=1:length(channels)
        if (isElementOf(channels(i).f,dataFlow)) then
            channel = channels(i);
            break
        end
    end
endfunction

// return the zone containing the given node
function zone = getZoneFromNode(node,zones)
    zone = -1;
    for i=1:length(zones)
        if (isElementOf(zones(i).n,node)) then
            zone = zones(i);
            break
        end
    end
endfunction

// if it exists, return the contiguity containing the channel whit the given data flow,
// otherwise return "-1"
function contiguity = getContFromDataFlow(dataFlow,nodes,zones,contiguities)
    contiguity = -1;
    srcTask = dataFlow.ts;
    dstTask = dataFlow.td;
    srcNode = getNodeFromTask(srcTask,nodes);
    dstNode = getNodeFromTask(dstTask,nodes);
    srcZone = getZoneFromNode(srcNode,zones);
    dstZone = getZoneFromNode(dstNode,zones);
    
    if ((srcZone <> -1) & (dstZone <> -1)) then
        for i=1:length(contiguities)
            if (((contiguities(i).z1.name == srcZone.name) & (contiguities(i).z2.name == dstZone.name)) | ((contiguities(i).z1.name == dstZone.name) & (contiguities(i).z2.name == srcZone.name)) & (srcZone.name <> dstZone.name)) then
                contiguity = contiguities(i);
                break
            end
        end
    end
    
endfunction
