///////////////////////////////
//  OPTIMIZATION STRATEGIES  //
///////////////////////////////

exec('optimizations/nwUtilities_functions.sci',0);
exec('optimizations/nwManipulations_functions.sci',0);

// ---------------
//  NODES REMOVAL
// ---------------
// Migrates all tasks, data flows and channels of the source node,
// to the destination node; then the source node is removed.

function x_neigh = nodesMerging(x_current, T, param)
    
    if (~isdef("param","local")) then
        param = [];
    end    
    
    step = x_current(3);
    
    network_init = x_current(1);
    network = x_current(2)
    manipulations_init = x_current(4);
    manipulations = x_current(5);
    
    // initialize current nodes and channels
    
    if (step == 0) then
        network = network_init;
        manipulations = manipulations_init;
    end
    
    // initialize new nodes and channels
    
    nodes = network.nodes;
    channels = network.abstractChannels;
    dataFlows = network.dataFlows;
    zones = network.zones;
    
    // check if there are more than 1 node
    
    if (length(nodes) > 1) then
        // candidate node to remove
        srcNode_index = int(rand()*length(nodes)+1);
        srcNode = nodes(srcNode_index);
        
        // choose randomly if try or not to replace the current node
        // with one in the technological library
        //
        if (round(rand()) & (length(techLib.nodes) <> 0)) then
            [isNodeReplaceable,nodes,channels,zones,manipulations,techLib] = replaceNode(srcNode_index,nodes,channels,zones,manipulations,techLib);
            if (isNodeReplaceable) then
                network.nodes = nodes;
                network.abstractChannels = channels;
                network.zones = zones;
                step = step+1;
                x_new = list(network_init,network,step,manipulations_init,manipulations,techLib);    
                x_neigh = nodesMerging(x_new, T, param);
            else
                step = 0;
                x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);
            end
            return
        end
        
        // candidate destination node
        dstNode_index = int(rand()*length(nodes)+1);
        dstNode = nodes(dstNode_index);
        
        // total capacity of tasks in the node to remove
        
        srcNode_capacity = [0 0];
        for i=1:length(srcNode.t)
            tasks = srcNode.t;
            t = tasks(i);
            srcNode_capacity = srcNode_capacity + t.c;
        end
        
        // remaining capacity of destination node
        
        dstNode_capacity = dstNode.c;
        for i=1:length(dstNode.t)
            tasks = dstNode.t;
            t = tasks(i);
            dstNode_capacity = dstNode_capacity - t.c;
        end
        
        // check if the node to remove contains at least one
        // task with mobility requirement
        
        srcTasksMobility = %f;
        for i=1:length(srcNode.t)
            tasks = srcNode.t;
            t = tasks(i);
            if (t.m == 1) then
                srcTasksMobility = %t;
                break
            end
        end
        
        // check if the destination node is mobile, then the source node
        // can be removed only if is connected with wireless channels.
        
        srcChannelsMobility = %t;
        if (dstNode.m == 1) then
            for i=1:length(channels)
                channel = channels(i);
                if ((channel.w == 0) & (isElementOf(channel.n,srcNode)) & (~isElementOf(channel.n,dstNode))) then
                    srcChannelsMobility = %f;
                    break
                end
            end            
        end

        // if possible, move tasks from source node to destination node
        // and remove source node and related channels
        
        if ((srcNode_capacity <= dstNode_capacity) & (dstNode_index <> srcNode_index) & srcChannelsMobility & (~srcTasksMobility | (srcTasksMobility & (dstNode.m == 1)))) then
            // remove data flows between tasks of the source node and tasks of the destination node
            // and move channels connected to source node to the destination node
            //
            for i=1:length(channels)
                channel = channels(i);
                for j=1:length(channels(i).f)
                    ts = channels(i).f(j).ts;
                    td = channels(i).f(j).td;
                    if (((isElementOf(nodes(srcNode_index).t,ts)) & (isElementOf(nodes(dstNode_index).t,td))) | ((isElementOf(nodes(srcNode_index).t,td)) & (isElementOf(nodes(dstNode_index).t,ts)))) then
                        manipulations.removeDataFlows($+1) = struct('dataFlow',channels(i).f(j).name,'channel',channels(i).name);
                        [dataFlows,channel] = removeDataFlow(channels(i).f(j),dataFlows,channel);
                    end
                end
                channels(i) = channel;
                if ((isElementOf(channels(i).n,srcNode)) & (~isElementOf(channels(i).n,dstNode))) then
                    manipulations.moveChannels($+1) = struct('channel',channels(i).name,'old_node',srcNode.name,'new_node',dstNode.name);
                    channels(i) = moveChannel(channels(i),nodes(srcNode_index),nodes(dstNode_index));
                end
            end
            
            // move tasks from source node to destination node
            //
            for i=1:length(nodes(srcNode_index).t)
                [srcNode,dstNode] = moveTask(nodes(srcNode_index).t(i),srcNode,dstNode);
                manipulations.moveTasks($+1) = struct('task',nodes(srcNode_index).t(i).name,'srcNode',srcNode.name,'dstNode',dstNode.name);
            end
            nodes(srcNode_index) = srcNode;
            nodes(dstNode_index) = dstNode;
            
            // remove the source node
            //
            zone = getZoneFromNode(srcNode,zones);
            [nodes,channels,zones] = removeNode(nodes(srcNode_index),nodes,channels,zones);
            if (zone <> -1) then
                manipulations.removeNodes($+1) = struct('node',srcNode.name,'zone',zone.name);
            else
                manipulations.removeNodes($+1) = struct('node',srcNode.name,'zone',"");
            end

            network.dataFlows = dataFlows;
            network.nodes = nodes;
            network.abstractChannels = channels;
            network.zones = zones;
            step = step+1;
            x_new = list(network_init,network,step,manipulations_init,manipulations);
            x_neigh = nodesMerging(x_new, T, param);
        else
            step = 0;
            x_neigh = list(network_init,network,step,manipulations_init,manipulations); 
        end
    else
        step = 0;
        x_neigh = list(network_init,network,step,manipulations_init,manipulations); 
    end
endfunction

// ------------------
//  CHANNELS REMOVAL
// ------------------
// Migrates all data flows and linked tasks of the source channel,
// to the destination channel and linked nodes (for tasks);
// then the source channel is removed.

function x_neigh = channelsMerging(x_current, T, param)
    
    if (~isdef("param","local")) then
        param = [];
    end    

    step = x_current(3);

    network_init = x_current(1);
    network = x_current(2)
    manipulations_init = x_current(4);
    manipulations = x_current(5);

    // initialize current nodes and channels
    
    if (step == 0) then
        network = network_init;
        manipulations = manipulations_init;
    end
    
    // initialize new nodes and channels
    
    nodes = network.nodes;
    channels = network.abstractChannels;
    dataFlows = network.dataFlows;
    zones = network.zones;
    
    // check if there are more than 1 channel
    
    if (length(channels) > 1) then
        
        // candidate channel to remove
        srcChannel_index = int(rand()*length(channels)+1);
        srcChannel = channels(srcChannel_index);
        
        // choose randomly if try or not to replace the current channel
        // with one in the technological library
        
        if (round(rand()) & (length(techLib.abstractChannels) <> 0)) then
            [isChannelReplaceable,channels,manipulations,techLib] = replaceChannel(srcChannel_index,channels,manipulations,techLib);
            if (isChannelReplaceable) then
                network.abstractChannels = channels;
                step = step+1;
                x_new = list(network_init,network,step,manipulations_init,manipulations,techLib);    
                x_neigh = channelsMerging(x_new, T, param);
            else
                step = 0;
                x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);   
            end
            return
        end
        
        // data flows of the source channel
        srcDataFlows = srcChannel.f;

        if (length(srcDataFlows) == 0) then
            step = 0;
            [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
            [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
            x_neigh = list(network_init,network,step,manipulations_init,manipulations);
        else            
            isChannelRemovable = %t;
                    
            for i=1:length(srcDataFlows)
                
                // data flow to move
                dataFlow = srcDataFlows(i);
    
                // tasks linked by the data flow
                srcTask = dataFlow.ts;
                dstTask = dataFlow.td;
                
                // candidate destination channel
                dstChannel_index = int(rand()*length(channels)+1);
                dstChannel = channels(dstChannel_index);
                
                // candidate nodes, linked by the channel, in which move the tasks, linked by the data flow
                srcNode_index = int(rand()*length(dstChannel.n)+1);
                srcNode = dstChannel.n(srcNode_index);
        
                dstNode_index = int(rand()*length(dstChannel.n)+1);
                dstNode = dstChannel.n(dstNode_index);
                
                for i=1:length(nodes)
                    if (nodes(i).name == srcNode.name) then
                        srcNode_index = i;
                        break
                    end
                end
                
                srcNode = nodes(srcNode_index);
                
                for i=1:length(nodes)
                    if (nodes(i).name == dstNode.name) then
                        dstNode_index = i;
                        break
                    end
                end
                
                dstNode = nodes(dstNode_index);
                
                // current nodes of the tasks
                srcNode_old = getNodeFromTask(srcTask,nodes);
               
                for i=1:length(nodes)
                    if (nodes(i).name == srcNode_old.name) then
                        srcNode_oldIndex = i;
                        break
                    end
                end
          
                dstNode_old = getNodeFromTask(dstTask,nodes);
        
                for i=1:length(nodes)
                    if (nodes(i).name == dstNode_old.name) then
                        dstNode_oldIndex = i;
                        break
                    end
                end
                
                // remaining capacities of the nodes
                srcNode_capacity = srcNode.c;
                for i=1:length(srcNode.t)
                    tasks = srcNode.t;
                    t = tasks(i);
                    srcNode_capacity = srcNode_capacity - t.c;
                end
                       
                dstNode_capacity = dstNode.c;
                for i=1:length(dstNode.t)
                    tasks = dstNode.t;
                    t = tasks(i);
                    dstNode_capacity = dstNode_capacity - t.c;
                end
                     
                // check tasks to nodes mobility
                isTaskMobilityOk = %t;
                
                if (srcTask.m == 1) & (srcNode.m == 0) then
                    isTaskMobilityOk = %f;
                end
                if (dstTask.m == 1) & (dstNode.m == 0) then
                    isTaskMobilityOk = %f;
                end
                
//                // check if data flow was already moved
//                dataFlowAlreadyMoved = %f;
//                
//                for i=1:length(manipulations.moveDataFlows)
//                    if (dataFlow.name == manipulations.moveDataFlows(i).dataFlow) then
//                        dataFlowAlreadyMoved = %t;
//                        break
//                    end
//                end
                
                if (((srcTask.c <= srcNode_capacity) | (srcNode_index == srcNode_oldIndex)) & ((dstTask.c <= dstNode_capacity) | (dstNode_index == dstNode_oldIndex)) & isTaskMobilityOk & isDataFlowMovable(dstChannel,dataFlow) & (dstChannel_index <> srcChannel_index) & (dstNode_index <> srcNode_index)) then 
               
                    network_old = network;
                    manipulations_old = manipulations;
                    
                    isSrcTaskMovable = %t;
                    isDstTaskMovable = %t;
                
                    [channels(srcChannel_index),channels(dstChannel_index)] = moveDataFlow(dataFlow,channels(srcChannel_index),channels(dstChannel_index));  
                    manipulations.moveDataFlows($+1) = struct('dataFlow',dataFlow.name,'srcChannel',srcChannel.name,'dstChannel',dstChannel.name); 
                                             
                    network.abstractChannels = channels;
                   
                    if (srcNode_index <> srcNode_oldIndex) then
                        [network,isSrcTaskMovable,manipulations] = tryToMoveTask(srcTask,srcNode_oldIndex,srcNode_index,network,manipulations,dstTask);
                        if (isSrcTaskMovable) then
                            isSrcTaskMovable = checkDataFlows(network.nodes,network.abstractChannels,network_init.dataFlows,isSrcTaskMovable);
                        end
                    end
        
                    if (dstNode_index <> dstNode_oldIndex) then
                        [network,isDstTaskMovable,manipulations] = tryToMoveTask(dstTask,dstNode_oldIndex,dstNode_index,network,manipulations,srcTask);
                        if (isDstTaskMovable) then
                            isDstTaskMovable = checkDataFlows(network.nodes,network.abstractChannels,network_init.dataFlows,isDstTaskMovable);
                        end
                    end
           
                    if (~isSrcTaskMovable | ~isDstTaskMovable) then
                        network = network_old;
                        manipulations = manipulations_old;
                        step = 0;
                        [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
                        [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
                        x_neigh = list(network_init,network,step,manipulations_init,manipulations);
                        isChannelRemovable = %f;
                        break
                    end 
                else
                    step = 0;
                    [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
                    [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
                    x_neigh = list(network_init,network,step,manipulations_init,manipulations);
                    isChannelRemovable = %f;
                    break
                end
            end
            
            if (isChannelRemovable) then
                channels = removeChannel(srcChannel,channels);
                manipulations.removeChannels($+1) = struct('channel',srcChannel.name);
                
                network.dataFlows = dataFlows;
                network.nodes = nodes;
                network.abstractChannels = channels;
                step = step+1;
                x_new = list(network_init,network,step,manipulations_init,manipulations);
                x_neigh = channelsMerging(x_new, T, param);
            end  
        end 
    else
        step = 0;
        [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
        [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
        x_neigh = list(network_init,network,step,manipulations_init,manipulations); 
    end
endfunction

// --------------------
//  TASKS REALLOCATION
// --------------------
// Migrates a task from a source node to a destination node,
// together with its linked data flows and related channels.

function x_neigh = tasksReallocation(x_current, T, param)

    if (~isdef("param","local")) then
        param = [];
    end    
    
    step = x_current(3);
    
    network_init = x_current(1);
    network = x_current(2)
    manipulations_init = x_current(4);
    manipulations = x_current(5);
    
    techLib = x_current(6);
        
    // initialize current nodes and channels
    
    if (step == 0) then
        network = network_init;
        manipulations = manipulations_init;
        
        for i=1:length(techLib.nodes)
            techLib.nodes(i).instances = 0;
        end
    end
    
    nodes = network.nodes;
    channels = network.abstractChannels;
    dataFlows = network.dataFlows;
    zones = network.zones;
    
    // check if there are more than 1 node
    
    if (length(nodes) > 1) then
        
        // candidate node containing the task to move
        srcNode_index = int(rand()*length(nodes)+1);
        srcNode = nodes(srcNode_index);
        
        // tasks of the source channel
        tasks = srcNode.t;
        
        if (length(tasks) == 0) then
            step = 0;
            [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
            [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
            x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);
        else
            // choose randomly if try or not to replace the current node
            // with one in the technological library
            //
            if (round(rand()) & (length(techLib.nodes) <> 0)) then
                [isNodeReplaceable,nodes,channels,zones,manipulations,techLib] = replaceNode(srcNode_index,nodes,channels,zones,manipulations,techLib);
                if (isNodeReplaceable) then
                    network.nodes = nodes;
                    network.abstractChannels = channels;
                    network.zones = zones;
                    step = step+1;
                    x_new = list(network_init,network,step,manipulations_init,manipulations,techLib);    
                    x_neigh = tasksReallocation(x_new, T, param);
                else
                    step = 0;
                    [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
                    [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
                    x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);
                end
                return
            end

            // candidate task to move
            task_index = int(rand()*length(srcNode.t)+1);
            task = srcNode.t(task_index);
            
            // candidate destination node
            dstNode_index = int(rand()*length(nodes)+1);
            dstNode = nodes(dstNode_index);
            
            // remaining capacity of destination node
            dstNode_capacity = dstNode.c;
            for i=1:length(dstNode.t)
                tasks = dstNode.t;
                t = tasks(i);
                dstNode_capacity = dstNode_capacity - t.c;
            end
            
//            // check if task was already moved
//            taskAlreadyMoved = %f;
//            
//            for i=1:length(manipulations.moveTasks)
//                if (task.name == manipulations.moveTasks(i).task) then
//                    taskAlreadyMoved = %t;
//                    break
//                end
//            end
            
//            if ((task.c <= dstNode_capacity) & (dstNode_index <> srcNode_index) & ~taskAlreadyMoved & (((task.m == 1) & (dstNode.m == 1)) | (task.m == 0))) then
            if ((task.c <= dstNode_capacity) & (dstNode_index <> srcNode_index) & (((task.m == 1) & (dstNode.m == 1)) | (task.m == 0))) then
                
                network_old = network;
                manipulations_old = manipulations;
                
                isTaskMovable = %t;
                
                [network,isTaskMovable,manipulations] = tryToMoveTask(task,srcNode_index,dstNode_index,network,manipulations);
                if (isTaskMovable) then
                    isTaskMovable = checkDataFlows(network.nodes,network.abstractChannels,network_init.dataFlows,isTaskMovable);
                end
                
                if (isTaskMovable) then
                    step = step+1;
                    x_new = list(network_init,network,step,manipulations_init,manipulations,techLib);    
                    x_neigh = tasksReallocation(x_new, T, param);
                else
                    network = network_old;
                    manipulations = manipulations_old;
                    step = 0;
                    [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
                    [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
                    x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);
                end
            else
                step = 0;
                [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
                [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
                x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);
            end
        end
    else
        step = 0;
        [network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes] = cleanNodes(network.nodes,network.abstractChannels,network.zones,manipulations.removeNodes);
        [network.abstractChannels,manipulations.removeChannels] = cleanChannels(network.abstractChannels,manipulations.removeChannels);
        x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib); 
    end
endfunction

// -------------------------
//  DATA FLOWS REALLOCATION
// -------------------------
// Migrates a data flow from a source channel to a destination channel,
// together with its linked tasks, other linked data flows and related channels.

function x_neigh = dataFlowsReallocation(x_current, T, param)

    if (~isdef("param","local")) then
        param = [];
    end    
    
    step = x_current(3);
    
    network_init = x_current(1);
    network = x_current(2)
    manipulations_init = x_current(4);
    manipulations = x_current(5);
    
    techLib = x_current(6);
    
    // initialize current network and manipulations
    
    if (step == 0) then
        network = network_init;
        manipulations = manipulations_init;
        
        for i=1:length(techLib.abstractChannels)
            techLib.abstractChannels(i).instances = 0;
        end
    end
    
    nodes = network.nodes;
    channels = network.abstractChannels;
    dataFlows = network.dataFlows;
    
    // check if there are more than 1 node

    if ((length(dataFlows) > 0) & (length(channels) >= 1)) then
        
        // candidate data flow to move
        dataFlow_index = int(rand()*length(dataFlows)+1);
        dataFlow = dataFlows(dataFlow_index);
        
        // tasks linked by the data flow
        srcTask = dataFlow.ts;
        dstTask = dataFlow.td;
    
        // current channel of the data flow
        channel = getChannelFromDataFlow(dataFlow,channels);
        
        for i=1:length(channels)
            if (channels(i).name == channel.name) then
                channel_index = i;
                break
            end
        end
        
        // choose randomly if try or not to replace the current channel
        // with one in the technological library
    
        if (round(rand()) & (length(techLib.abstractChannels) <> 0)) then
            [isChannelReplaceable,channels,manipulations,techLib] = replaceChannel(channel_index,channels,manipulations,techLib);
            if (isChannelReplaceable) then
                network.abstractChannels = channels;
                step = step+1;
                x_new = list(network_init,network,step,manipulations_init,manipulations,techLib);    
                x_neigh = dataFlowsReallocation(x_new, T, param);
            else
                step = 0;
                x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);   
            end
            return
        end
        
        // candidate channel in which move the data flow   
        newChannel_index = int(rand()*length(channels)+1);
        newChannel = channels(newChannel_index);
        
        // candidate nodes, linked by the channel, in which move the tasks, linked by the data flow
        srcNode_index = int(rand()*length(newChannel.n)+1);
        srcNode = newChannel.n(srcNode_index);

        dstNode_index = int(rand()*length(newChannel.n)+1);
        dstNode = newChannel.n(dstNode_index);
        
        for i=1:length(nodes)
            if (nodes(i).name == srcNode.name) then
                srcNode_index = i;
                break
            end
        end
        
        srcNode = nodes(srcNode_index);
        
        for i=1:length(nodes)
            if (nodes(i).name == dstNode.name) then
                dstNode_index = i;
                break
            end
        end
        
        dstNode = nodes(dstNode_index);

        // current nodes of the tasks
        srcNode_old = getNodeFromTask(srcTask,nodes);
       
        for i=1:length(nodes)
            if (nodes(i).name == srcNode_old.name) then
                srcNode_oldIndex = i;
                break
            end
        end
  
        dstNode_old = getNodeFromTask(dstTask,nodes);

        for i=1:length(nodes)
            if (nodes(i).name == dstNode_old.name) then
                dstNode_oldIndex = i;
                break
            end
        end
        
        // remaining capacities of the nodes
        srcNode_capacity = srcNode.c;
        for i=1:length(srcNode.t)
            tasks = srcNode.t;
            t = tasks(i);
            srcNode_capacity = srcNode_capacity - t.c;
        end
               
        dstNode_capacity = dstNode.c;
        for i=1:length(dstNode.t)
            tasks = dstNode.t;
            t = tasks(i);
            dstNode_capacity = dstNode_capacity - t.c;
        end
             
        // check tasks to nodes mobility
        isTaskMobilityOk = %t;
        
        if (srcTask.m == 1) & (srcNode.m == 0) then
            isTaskMobilityOk = %f;
        end
        if (dstTask.m == 1) & (dstNode.m == 0) then
            isTaskMobilityOk = %f;
        end
        
        
        
//        // check if tasks were already moved
//        taskAlreadyMoved = %f;
//        
//        for i=1:length(manipulations.moveTasks)
//            if ((srcTask.name == manipulations.moveTasks(i).task) | (dstTask.name == manipulations.moveTasks(i).task)) then
//                taskAlreadyMoved = %t;
//                break
//            end
//        end

//        // check if data flow was already moved
//        dataFlowAlreadyMoved = %f;
//        
//        for i=1:length(manipulations.moveDataFlows)
//            if (dataFlow.name == manipulations.moveDataFlows(i).dataFlow) then
//                dataFlowAlreadyMoved = %t;
//                break
//            end
//        end

        if (((srcTask.c <= srcNode_capacity) | (srcNode_index == srcNode_oldIndex)) & ((dstTask.c <= dstNode_capacity) | (dstNode_index == dstNode_oldIndex)) & isTaskMobilityOk & isDataFlowMovable(newChannel,dataFlow) & (newChannel_index <> channel_index) & (dstNode_index <> srcNode_index)) then 
            
            network_old = network;
            manipulations_old = manipulations;
            
            isSrcTaskMovable = %t;
            isDstTaskMovable = %t;
        
            [channels(channel_index),channels(newChannel_index)] = moveDataFlow(dataFlow,channels(channel_index),channels(newChannel_index));  
            manipulations.moveDataFlows($+1) = struct('dataFlow',dataFlow.name,'srcChannel',channel.name,'dstChannel',newChannel.name);      
            network.abstractChannels = channels;
           
            if (srcNode_index <> srcNode_oldIndex) then
                [network,isSrcTaskMovable,manipulations] = tryToMoveTask(srcTask,srcNode_oldIndex,srcNode_index,network,manipulations,dstTask);
                if (isSrcTaskMovable) then
                    isSrcTaskMovable = checkDataFlows(network.nodes,network.abstractChannels,network_init.dataFlows,isSrcTaskMovable);
                end
            end

            if (dstNode_index <> dstNode_oldIndex) then
                [network,isDstTaskMovable,manipulations] = tryToMoveTask(dstTask,dstNode_oldIndex,dstNode_index,network,manipulations,srcTask);
                if (isDstTaskMovable) then
                    isDstTaskMovable = checkDataFlows(network.nodes,network.abstractChannels,network_init.dataFlows,isDstTaskMovable);
                end
            end
            
            if (isSrcTaskMovable & isDstTaskMovable) then
                step = step+1;
                x_new = list(network_init,network,step,manipulations_init,manipulations,techLib);    
                x_neigh = dataFlowsReallocation(x_new, T, param);
            else
                network = network_old;
                manipulations = manipulations_old;
                step = 0;
                x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);
            end        
        else
            step = 0;
            x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);
        end
    else
        step = 0;
        x_neigh = list(network_init,network,step,manipulations_init,manipulations,techLib);
    end
endfunction

/////////////////
//  FUNCTIONS  //
/////////////////

function isMovable = isDataFlowMovable(channel, dataFlow)
    remainingThroughput = channel.c.throughput;

    for i=1:length(channel.f)
        dataFlows = channel.f;
        df = dataFlows(i);
        remainingThroughput = remainingThroughput - df.c.max_throughput;
    end

    if ((dataFlow.c.max_throughput <= remainingThroughput) & (dataFlow.c.max_delay >= channel.c.delay) & (dataFlow.c.max_error_rate >= channel.c.error_rate)) then
        isMovable = %t;
    else
        isMovable = %f;
    end
endfunction

// checks whether the current data flows are consistent with the initial
function areDataFlowsOk = checkDataFlows(nodes, channels, dataFlows_init, areDataFlowsOk)
    for i=1:length(dataFlows_init)
        ts = dataFlows_init(i).ts;
        td = dataFlows_init(i).td;
        ok = %f;
        for j=1:length(channels)
            if (isElementOf(channels(j).f,dataFlows_init(i))) then
                ok = %t;
                break
            end
        end
        
        if (~ok) then
            for j=1:length(nodes)
                if ((isElementOf(nodes(j).t,ts)) & (isElementOf(nodes(j).t,td))) then
                    ok = %t;
                    break
                end
            end

            if (~ok) then
                areDataFlowsOk = %f;
                break
            end
        end
    end
endfunction

// check if the possible data flows related to the task can be moved to the new node.
// There are two alternatives:
// - the channels containing this data flows contain ONLY them;
// - otherwise, must exist a channel between the involved nodes  with enough capacity to contain the data flows.
function [network,isTaskMovable,manipulations] = tryToMoveTask(srcTask, srcNode_index, dstNode_index, network, manipulations,dstTask)
    isTaskMovable = %t;
    
    nodes = network.nodes;
    channels = network.abstractChannels;
    dataFlows = network.dataFlows;
    
    srcNode = nodes(srcNode_index);
    dstNode = nodes(dstNode_index);
    
    moveTasks = manipulations.moveTasks;
    removeDataFlows = manipulations.removeDataFlows;
    moveDataFlows = manipulations.moveDataFlows;
    moveChannels = manipulations.moveChannels;
    
    linkedDataFlows = list();
    movedDataFlows = list();
    
    for i=1:length(channels)
        for j=1:length(channels(i).f)
            if (channels(i).f(j).ts.name == srcTask.name) then
                ts = channels(i).f(j).ts;
                td = channels(i).f(j).td;
            elseif (channels(i).f(j).td.name == srcTask.name) then
                td = channels(i).f(j).ts;
                ts = channels(i).f(j).td;
            else
                continue
            end       
            
            if (~((isdef("dstTask","local")) & (td.name == dstTask.name))) then
                linkedDataFlows($+1) = struct('linked_task',td,'channel_index',i,'dataFlow_index',j);
            end
        end
    end
    
    for i=1:length(linkedDataFlows)
        t = linkedDataFlows(i).linked_task;

        tNode = getNodeFromTask(t,nodes);
        ch_index = linkedDataFlows(i).channel_index;
        df_index = linkedDataFlows(i).dataFlow_index;
        
        // the linked task belongs to the destination node
        // --> remove the data flow between the involved tasks
        //
        if (isElementOf(dstNode.t,t)) then
            removeDataFlows($+1) = struct('dataFlow',channels(ch_index).f(df_index).name,'channel',channels(ch_index).name);
            [dataFlows,channels(ch_index)] = removeDataFlow(channels(ch_index).f(df_index),dataFlows,channels(ch_index));
        //
        // the channel, containing the data flow between the involved tasks, contains only the involved data flow
        // --> move the channel to the node containing the linked task
        //
        elseif (length(channels(ch_index).f) == 1) then 
            if (((dstNode.m == 1) & (channels(ch_index).w == 1)) | (dstNode.m == 0)) then
                moveChannels($+1) = struct('channel',channels(ch_index).name,'old_node',srcNode.name,'new_node',dstNode.name);
                channels(ch_index) = moveChannel(channels(ch_index),srcNode,dstNode);
            else
                isTaskMovable = %f;
            end
        //
        // otherwise, if exists, try to find a channel between the nodes containing the involved tasks,
        // which can hold the data flow between the involved tasks
        // --> move the data flow between the involved tasks, in the new channel
        // --> if the channel can't contain the data flow, returns a false boolean
        //
        else
            dstChannel_index = int(rand()*length(channels)+1);

            if ((isDataFlowMovable(channels(dstChannel_index),channels(ch_index).f(df_index))) & (dstChannel_index <> ch_index) & (isElementOf(channels(dstChannel_index).n,dstNode)) & (isElementOf(channels(dstChannel_index).n,tNode))) then
                moveDataFlows($+1) = struct('dataFlow',channels(ch_index).f(df_index).name,'srcChannel',channels(ch_index).name,'dstChannel',channels(dstChannel_index).name);
                movedDataFlows($+1) = channels(ch_index).f(df_index);
                [channels(ch_index),channels(dstChannel_index)] = moveDataFlow(channels(ch_index).f(df_index),channels(ch_index),channels(dstChannel_index));
            elseif (~isElementOf(movedDataFlows,channels(ch_index).f(df_index))) then
                isTaskMovable = %f;
                break
            end
        end
    end
    
    if (isTaskMovable) then
        [nodes(srcNode_index),nodes(dstNode_index)] = moveTask(srcTask,nodes(srcNode_index),nodes(dstNode_index));
        moveTasks($+1) = struct('task',srcTask.name,'srcNode',srcNode.name,'dstNode',dstNode.name);
    end
    
    network.nodes = nodes;
    network.abstractChannels = channels;
    network.dataFlows = dataFlows;
    
    manipulations.moveTasks = moveTasks;
    manipulations.removeDataFlows = removeDataFlows;
    manipulations.moveDataFlows = moveDataFlows;
    manipulations.moveChannels = moveChannels;
endfunction

function [isNodeReplaceable,nodes,channels,zones,manipulations,techLib] = replaceNode(srcNode_index,nodes,channels,zones,manipulations,techLib)

    srcNode = nodes(srcNode_index);

    // candidate node from the technological library
    // to replace the current node
    
    nodesLib = techLib.nodes;
    dstNode_index = int(rand()*length(nodesLib)+1);            
    dstNode = nodesLib(dstNode_index);
    dstNode_libName = dstNode.name;
    
    isMobitityOk = %t;
    totalCapacity = [0 0];
    
    for i=1:length(srcNode.t)
    if ((srcNode.t(i).m == 1) & (dstNode.m == 0)) then
        isMobilityOk = %f;
        break
    end
    totalCapacity = totalCapacity + srcNode.t(i).c;
    end
    
    if ((isMobitityOk) & (totalCapacity <= dstNode.c)) then
        isNodeReplaceable = %t;
        techLib.nodes(dstNode_index).instances = techLib.nodes(dstNode_index).instances + 1;
        dstNode.name = dstNode.name + "_" + string(techLib.nodes(dstNode_index).instances);
        
        // move tasks from source node to destination node
        //
        for i=1:length(nodes(srcNode_index).t)
            [srcNode,dstNode] = moveTask(nodes(srcNode_index).t(i),srcNode,dstNode);
            manipulations.moveTasks($+1) = struct('task',nodes(srcNode_index).t(i).name,'srcNode',srcNode.name,'dstNode',dstNode.name);
        end
        nodes(srcNode_index) = srcNode;
        
        // move channels from source node to destination node
        //
        for i=1:length(channels)
            if (isElementOf(channels(i).n,srcNode)) then
                channels(i) = moveChannel(channels(i),srcNode,dstNode);
                manipulations.moveChannels($+1) = struct('channel',channels(i).name,'old_node',srcNode.name,'new_node',dstNode.name);
            end
        end
        
        // remove the source node
        //
        srcZone = getZoneFromNode(srcNode,zones);
        [nodes,channels,zones] = removeNode(nodes(srcNode_index),nodes,channels,zones);
        if (srcZone <> -1) then
            manipulations.removeNodes($+1) = struct('node',srcNode.name,'zone',srcZone.name);
        else
            manipulations.removeNodes($+1) = struct('node',srcNode.name,'zone',"");
        end
        
        // add the destination node to the network
        //                 
        [nodes,zones] = addNode(dstNode,srcZone,nodes,zones);
        if (srcZone <> -1) then
            manipulations.addNodes($+1) = struct('libNode',dstNode_libName,'node',dstNode.name,'zone',srcZone.name);
        else
            manipulations.addNodes($+1) = struct('libNode',dstNode_libName,'node',dstNode.name,'zone',"");
        end
    else
        isNodeReplaceable = %f;
    end
endfunction

function [isChannelReplaceable,channels,manipulations,techLib] = replaceChannel(channel_index,channels,manipulations,techLib)
    
    channel = channels(channel_index);
    
    // candidate channel from the technological library
    // to replace the current channel
    
    channelsLib = techLib.abstractChannels;
    newChannel_index = int(rand()*length(channelsLib)+1);         
    newChannel = channelsLib(newChannel_index);
    newChannel_libName = newChannel.name;
              
    if ((channel.w == 0) | ((channel.w == 1) & (newChannel.w == 1))) then
        isCapacityOk = %t;
        totalThroughput = 0;
        for i=1:length(channel.f)
            if (channel.f(i).c.max_delay < newChannel.c.delay) then
                isCapacityOk = %f;
                break
            end
            if (channel.f(i).c.max_error_rate < newChannel.c.error_rate) then
                isCapacityOk = %f;
                break
            end
            totalThroughput = totalThroughput + channel.f(i).c.max_throughput;
        end
        if (isCapacityOk & (totalThroughput > newChannel.c.throughput)) then
            isCapacityOk = %f;
        end
        
        if (isCapacityOk) then
            techLib.abstractChannels(newChannel_index).instances = techLib.abstractChannels(newChannel_index).instances + 1;
            newChannel.name = newChannel.name + "_" + string(techLib.abstractChannels(newChannel_index).instances);
            channels = addChannel(newChannel,channels);
            manipulations.addChannels($+1) = struct('libChannel',newChannel_libName,'channel',newChannel.name);
            newChannel_index = length(channels);

            for i=1:length(channel.f)
                dataFlow = channel.f(i);
                [channels(channel_index),channels(newChannel_index)] = moveDataFlow(dataFlow,channels(channel_index),channels(newChannel_index));  
                manipulations.moveDataFlows($+1) = struct('dataFlow',dataFlow.name,'srcChannel',channel.name,'dstChannel',newChannel.name);
            end
            
            for i=1:length(channel.n)
                node = channel.n(i);
                fakeNode = struct('name',"");
                channels(newChannel_index) = moveChannel(channels(newChannel_index),fakeNode,node);
                manipulations.moveChannels($+1) = struct('channel',channels(newChannel_index).name,'old_node',fakeNode.name,'new_node',node.name);
            end
            
            channels = removeChannel(channel,channels);
            manipulations.removeChannels($+1) = struct('channel',channel.name);
            
            isChannelReplaceable = %t;
        else
            isChannelReplaceable = %f;
        end
    else
        isChannelReplaceable = %f;
    end
endfunction
